#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>

#include "stdbool.h"	
#include "spi.h"
#include "ST7735R_TFT.h"
#include "usart.h"

#define LEFT 0
#define RIGHT 1
#define DOWN 2
#define UP 3

#define UP_BUTTON (1 << PC6)
#define LEFT_BUTTON (1 << PC0)
#define RIGHT_BUTTON (1 << PC2)
#define DOWN_BUTTON (1 << PC4)

#define UP_PRESSED ((PINC & UP_BUTTON) == 0)
#define DOWN_PRESSED ((PINC & DOWN_BUTTON) == 0)
#define LEFT_PRESSED ((PINC & LEFT_BUTTON) == 0)
#define RIGHT_PRESSED ((PINC & RIGHT_BUTTON) == 0)

int16_t dim_max_sarpe = 60;
int16_t dim_sarpe = 10;
int16_t dim_init = 10;
int16_t speed = 1;
int16_t food_elems = 3;
int16_t available_food;
int16_t level;
int32_t directie;

char ok = 1;
int score;

typedef struct {
	int16_t x;
	int16_t y;
} coord;

coord coord_sarpe[60], coord_mancare[30];

void init_sarpte(void) {
	int16_t i;
	for (i = 0; i < dim_init; ++i) {
		coord_sarpe[i].x = 10;
		coord_sarpe[i].y = ST7735R_WIDTH - i - 1;
	}
}

void afisare_ecran_start(void) {
	char buf[32];
	
	ST7735R_FillRect(0, 0, ST7735R_WIDTH + 10, ST7735R_HEIGHT + 20, 0, 0, 0);
	sprintf(buf, "SNAKE RELOADED");
	ST7735R_DrawText(20, 60, buf, 255, 255, 255, 0, 0, 0);
	sprintf(buf, "Apasati orice");
	ST7735R_DrawText(30, 90, buf, 255, 255, 255, 0, 0, 0);
	sprintf(buf, "\t\t\t\tbuton");
	ST7735R_DrawText(30, 105, buf, 255, 255, 255, 0, 0, 0);
	sprintf(buf, "pentru a incepe");
	ST7735R_DrawText(25, 120, buf, 255, 255, 255, 0, 0, 0);
	
	_delay_ms(500);
}

void afisare_win_game(void) {
	char buf[32];

	ST7735R_FillRect(0, 0, ST7735R_WIDTH + 10, ST7735R_HEIGHT + 20, 0, 0, 0);
	sprintf(buf, "YOU WIN");
	ST7735R_DrawText(30, 90, buf, 255, 255, 255, 0, 0, 0);	
	sprintf(buf, "Score: %d", score);
	ST7735R_DrawText(40, 110, buf, 255, 255, 255, 0, 0, 0);
	sprintf(buf, "Level: 6");
	ST7735R_DrawText(40, 130, buf, 255, 255, 255, 0, 0, 0);
}

void initializare_butoane(void) {
	// activeaza rezistentele de pullup
	PORTC |= (1 << PC0);
	PORTC |= (1 << PC2);
	PORTC |= (1 << PC4);
	PORTC |= (1 << PC6);

	DDRC &= ~(1 << PC0);
	DDRC &= ~(1 << PC2);
	DDRC &= ~(1 << PC4);
	DDRC &= ~(1 << PC6);
}

void seteaza_directie_mers(void) {
	if (DOWN_PRESSED && directie != UP) directie = DOWN;
	else if (RIGHT_PRESSED && directie != LEFT) directie = RIGHT;
	else if (LEFT_PRESSED && directie  != RIGHT) directie = LEFT;
	else if (UP_PRESSED && directie != DOWN) directie = UP;
}

void afisare_ecran_final(void) {
	char buf[32];
	
	ST7735R_FillRect(0, 0, ST7735R_WIDTH + 10, ST7735R_HEIGHT + 20, 0, 0, 0);
	sprintf(buf, "Game over");
	ST7735R_DrawText(30, 90, buf, 255, 255, 255, 0, 0, 0);
	sprintf(buf, "Score: %d", score);
	ST7735R_DrawText(40, 110, buf, 255, 255, 255, 0, 0, 0);
	sprintf(buf, "Level: %d", level);
	ST7735R_DrawText(40, 130, buf, 255, 255, 255, 0, 0, 0);
}

void level_up(void) {
	char buf[15];
	ST7735R_FillRect(0, 0, ST7735R_WIDTH + 10, ST7735R_HEIGHT + 20, 0, 0, 0);
	sprintf(buf, "Level up!");
	ST7735R_DrawText(25, 100, buf, 255, 255, 255, 0, 0, 0);
	sprintf(buf, "Nivel: %d", level);
	ST7735R_DrawText(25, 120, buf, 255, 255, 255, 0, 0, 0);
}

bool verifica_final_joc(void) {
	int16_t i;
	// verifica coliziunile cu propriul corp
	for (i = 1; i < dim_sarpe; ++i) {
		if (coord_sarpe[0].x == coord_sarpe[i].x && coord_sarpe[0].y == coord_sarpe[i].y) {
			return true;
		}
	}

	// verifica coliziunile cu peretii imaginari:
	if (coord_sarpe[dim_sarpe - 1].x == 0) return true;
	if (coord_sarpe[dim_sarpe - 1].x == ST7735R_WIDTH) return true;
	if (coord_sarpe[dim_sarpe - 1].y == 0) return true;	
	if (coord_sarpe[dim_sarpe - 1].y == ST7735R_HEIGHT) return true;	

	return false;
}

void deseneaza_sarpe(void) {
	int16_t i;
	for (i = 0; i < dim_sarpe; ++i) {
		ST7735R_FillRect(coord_sarpe[i].x, coord_sarpe[i].y, 
			coord_sarpe[i].x + 1, coord_sarpe[i].y + 1, 255, 255, 0);
	}
	
	for (i = 0; i < food_elems; ++i) {
		ST7735R_FillRect(coord_mancare[i].x,coord_mancare[i].y,
		coord_mancare[i].x + 1, coord_mancare[i].y + 1, 255, 120, 125);
	}
	
	_delay_ms(200);
}

void mancare_random(void) {
	int16_t i;
	for (i = 0; i < food_elems; ++i) {
		coord_mancare[i].x = rand() % 127;
		coord_mancare[i].y = rand() % 159;
	}
	available_food = food_elems;
}

void mananca(void) {
	int16_t i;
	for (i = 0; i < food_elems; ++i) {
		if (coord_mancare[i].x == -1) continue;
		if (coord_mancare[i].x == coord_sarpe[0].x 
			&& coord_sarpe[0].y == coord_mancare[i].y) {
			++score;
			++dim_sarpe;
			coord_sarpe[dim_sarpe - 1].x = coord_sarpe[dim_sarpe - 2].x;
			coord_sarpe[dim_sarpe - 1].y = coord_sarpe[dim_sarpe - 2].y;
			--available_food;

			break;
		}	
	}
	
	if (dim_sarpe == dim_max_sarpe) {
		afisare_win_game();
	}
}

void mareste_viteza(void) {
	if (ok == 1 && score % 10 == 0) {
		ok = 0;
		++level;
		++speed;
		food_elems += 3;
		level_up();
	}
}

void modifica_coord_sarpe(int32_t dir) {
	int16_t i;

	switch(dir) {
		case LEFT:
			++coord_sarpe[dim_sarpe - 1].y;
			break;
		case UP:
			--coord_sarpe[dim_sarpe - 1].x;
			break;
		case RIGHT:
			--coord_sarpe[dim_sarpe - 1].y;
			break;
		case DOWN:
			++coord_sarpe[dim_sarpe - 1].x;
			break;
	}

	for (i = 1; i < dim_sarpe; ++i) {
		coord_sarpe[i - 1].x = coord_sarpe[i].x;
		coord_sarpe[i - 1].y = coord_sarpe[i].y;
	}	

}

int main(void) {
	directie = DOWN;
	
	initializare_butoane();
	USART0_init();
	SPI_init();
	ST7735R_Begin();
	ST7735R_FillRect(0, 0, ST7735R_WIDTH + 10, ST7735R_HEIGHT + 20, 0, 0, 0);

	afisare_ecran_start();
	init_sarpte();
	
	if (RIGHT_PRESSED || LEFT_PRESSED || UP_PRESSED || DOWN_PRESSED) {
		while (1) {
			ST7735R_FillRect(0, 0, ST7735R_WIDTH + 10, ST7735R_HEIGHT + 20, 0, 0, 0);
			if (available_food == 0) {
				mancare_random();
			}
			mareste_viteza();
			seteaza_directie_mers();
			modifica_coord_sarpe(directie);
			ST7735R_FillRect(0, 0, ST7735R_WIDTH + 10, ST7735R_HEIGHT + 20, 0, 0, 0);
			deseneaza_sarpe();
			mananca();

			if (verifica_final_joc() == true) {
				afisare_ecran_final();
				break;
			}

			_delay_ms(100);
		}
	}
	
	return 0;
}